import cv2
import numpy as np

import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.preprocessImage import preprocessImage


class GetImagesMixin:
  def _getImage(self, frameNumber, wellNumber):
    minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
    debug = 0

    xtop = self._wellPositions[wellNumber]['topLeftX']
    ytop = self._wellPositions[wellNumber]['topLeftY']
    lenX = self._wellPositions[wellNumber]['lengthX']
    lenY = self._wellPositions[wellNumber]['lengthY']

    ret = False
    while (not(ret)):
      cap = zzVideoReading.VideoCapture(self._videoPath)
      cap.set(1, frameNumber)
      ret, frame = cap.read()
      if not(ret):
        frameNumber = frameNumber - 1

    if self._hyperparameters["invertBlackWhiteOnImages"]:
      frame = 255 - frame

    if self._hyperparameters["imagePreProcessMethod"]:
      frame = preprocessImage(frame, self._hyperparameters)

    grey = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]

    if (debug):
      self._debugFrame(curFrame, title='Frame')

    return curFrame

  def _getForegroundImageSequential(self, cap, frameNumber, wellNumber, alreadyExtractedImage=0, trackingHeadTailAllAnimals=0):
    minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
    if "minPixelDiffForBackExtractHead" in self._hyperparameters:
      minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtractHead"]
    debug = 0

    xtop = self._wellPositions[wellNumber]['topLeftX']
    ytop = self._wellPositions[wellNumber]['topLeftY']
    lenX = self._wellPositions[wellNumber]['lengthX']
    lenY = self._wellPositions[wellNumber]['lengthY']

    back = self._background[ytop:ytop+lenY, xtop:xtop+lenX]

    if (type(alreadyExtractedImage) != int):
      ret   = True
      frame = alreadyExtractedImage.copy()
    else:
      ret, frame = cap.read()

    if not(ret):
      if self._hyperparameters["searchPreviousFramesIfCurrentFrameIsCorrupted"]:
        currentFrameNum = int(cap.get(1))
        while not(ret) and currentFrameNum:
          currentFrameNum = currentFrameNum - 1
          cap.set(1, currentFrameNum)
          ret, frame = cap.read()
      else:
        frame = back.copy()

    if self._hyperparameters["invertBlackWhiteOnImages"]:
      frame = 255 - frame

    if self._hyperparameters["imagePreProcessMethod"]:
      frame = preprocessImage(frame, self._hyperparameters)

    grey = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]

    xHead = trackingHeadTailAllAnimals[0][frameNumber - self._hyperparameters["firstFrame"] - 1][0][0]
    yHead = trackingHeadTailAllAnimals[0][frameNumber - self._hyperparameters["firstFrame"] - 1][0][1]
    if self._hyperparameters["trackOnlyOnROI_halfDiameter"] != 0 and frameNumber != self._hyperparameters["firstFrame"] and xHead != 0 and yHead != 0:
      xHeadMin = xHead
      xHeadMax = xHead
      yHeadMin = yHead
      yHeadMax = yHead
      for animalId in range(1, self._hyperparameters["nbAnimalsPerWell"]):
        xHead = trackingHeadTailAllAnimals[animalId][frameNumber - self._hyperparameters["firstFrame"] - 1][0][0]
        yHead = trackingHeadTailAllAnimals[animalId][frameNumber - self._hyperparameters["firstFrame"] - 1][0][1]
        xHeadMin = xHeadMin if xHeadMin < xHead else xHead
        xHeadMax = xHeadMax if xHeadMax > xHead else xHead
        yHeadMin = yHeadMin if yHeadMin < yHead else yHead
        yHeadMax = yHeadMax if yHeadMax > yHead else yHead
      maxHalfDiameter = self._hyperparameters["trackOnlyOnROI_halfDiameter"]
      xmin = int(xHeadMin - maxHalfDiameter) if xHeadMin - maxHalfDiameter >= 0 else 0
      ymin = int(yHeadMin - maxHalfDiameter) if yHeadMin - maxHalfDiameter >= 0 else 0
      xmax = int(xHeadMax + maxHalfDiameter) if xHeadMax + maxHalfDiameter < len(curFrame[0]) else len(curFrame[0]) - 1
      ymax = int(yHeadMax + maxHalfDiameter) if yHeadMax + maxHalfDiameter < len(curFrame)    else len(curFrame) - 1
      curFrameInitial = curFrame.copy()
      backInitial     = back.copy()
      initialCurFrameInitial = curFrame.copy()
      initialCurFrame = curFrame.copy()
      curFrame = curFrame[ymin:ymax, xmin:xmax]
      back     = back[ymin:ymax, xmin:xmax]
    else:
      xmin = 0
      ymin = 0
      initialCurFrameInitial = 0
      curFrameInitial = 0
      backInitial     = 0
      initialCurFrame = curFrame.copy()

    putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )

    curFrame[putToWhite] = 255
    previousNbBlackPixels = []
    if self._hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"]:
      minPixel2nbBlackPixels = {}
      countTries = 0
      nbBlackPixels = 0
      nbBlackPixelsMax = int(self._hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"])
      while (minPixelDiffForBackExtract > 0) and (countTries < 30) and not(minPixelDiffForBackExtract in minPixel2nbBlackPixels):
        if countTries > 0:
          curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]
          if self._hyperparameters["trackOnlyOnROI_halfDiameter"] != 0 and frameNumber != self._hyperparameters["firstFrame"] and xHead != 0 and yHead != 0:
            curFrame = curFrame[ymin:ymax, xmin:xmax]
          putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
          curFrame[putToWhite] = 255
        ret, thresh1 = cv2.threshold(curFrame, self._hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
        thresh1 = 255 - thresh1
        nbBlackPixels = cv2.countNonZero(thresh1)
        minPixel2nbBlackPixels[minPixelDiffForBackExtract] = nbBlackPixels
        if nbBlackPixels > nbBlackPixelsMax:
          minPixelDiffForBackExtract = minPixelDiffForBackExtract + 1
        if nbBlackPixels <= nbBlackPixelsMax:
          minPixelDiffForBackExtract = minPixelDiffForBackExtract - 1
        countTries = countTries + 1
        previousNbBlackPixels.append(nbBlackPixels)
        if len(previousNbBlackPixels) >= 3:
          lastThree = previousNbBlackPixels[len(previousNbBlackPixels)-3: len(previousNbBlackPixels)]
          if lastThree.count(lastThree[0]) == len(lastThree):
            countTries = 1000000

      best_minPixelDiffForBackExtract = 0
      minDist = 10000000000000
      for minPixelDiffForBackExtract in minPixel2nbBlackPixels:
        nbBlackPixels = minPixel2nbBlackPixels[minPixelDiffForBackExtract]
        dist = abs(nbBlackPixels - nbBlackPixelsMax)
        if dist < minDist:
          minDist = dist
          best_minPixelDiffForBackExtract = minPixelDiffForBackExtract

      minPixelDiffForBackExtract = best_minPixelDiffForBackExtract
      self._hyperparameters["minPixelDiffForBackExtractHead"] = minPixelDiffForBackExtract
      curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]
      if self._hyperparameters["trackOnlyOnROI_halfDiameter"] != 0 and frameNumber != self._hyperparameters["firstFrame"] and xHead != 0 and yHead != 0:
        curFrame = curFrame[ymin:ymax, xmin:xmax]
        curFrameInitial = curFrame
        backInitial     = back
        # initialCurFrameInitial = curFrame
      putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
      curFrame[putToWhite] = 255

    # else:
      # putToBlack3a = ( curFrame.astype('int32') <= (back.astype('int32') - minPixelDiffForBackExtract) )
      # putToBlack3b = ( curFrame.astype('int32') >= (back.astype('int32') + minPixelDiffForBackExtract) )

      # putToWhite  = (curFrame <= 220)
      # putToBlack  = (curFrame >= 220)

      # curFrame[putToWhite]   = 255
      # curFrame[putToBlack]   = 0
      # curFrame[putToBlack3a] = 0
      # curFrame[putToBlack3b] = 0

    if debug:
      self._debugFrame(frame, title='Frame')

    if self._hyperparameters["trackOnlyOnROI_halfDiameter"] != 0:
      ret, res = cv2.threshold(curFrame, self._hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
      if ret:
        reinitialize = 1
        contours, hierarchy = cv2.findContours(res, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        for contour in contours:
          contourArea = cv2.contourArea(contour)
          if contourArea > self._hyperparameters["minAreaBody"] and contourArea < self._hyperparameters["maxAreaBody"]:
            reinitialize = 0
        if reinitialize and type(curFrameInitial) != int:
          curFrame = curFrameInitial
          initialCurFrame = initialCurFrameInitial
          back = backInitial

    return [curFrame, initialCurFrame, back, xmin, ymin]

  def getForegroundImage(self, background, frameNumber, wellNumber):
    minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
    if "minPixelDiffForBackExtractHead" in self._hyperparameters:
      minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtractHead"]
    debug = 0

    if len(self._wellPositions):
      xtop = self._wellPositions[wellNumber]['topLeftX']
      ytop = self._wellPositions[wellNumber]['topLeftY']
      lenX = self._wellPositions[wellNumber]['lengthX']
      lenY = self._wellPositions[wellNumber]['lengthY']
    else:
      xtop = 0
      ytop = 0
      lenX = len(background[0])
      lenY = len(background)

    back = background[ytop:ytop+lenY, xtop:xtop+lenX]

    cap = zzVideoReading.VideoCapture(self._videoPath)
    cap.set(1, frameNumber)
    ret, frame = cap.read()

    if not(ret):
      if self._hyperparameters["searchPreviousFramesIfCurrentFrameIsCorrupted"]:
        currentFrameNum = int(cap.get(1))
        while not(ret) and currentFrameNum:
          currentFrameNum = currentFrameNum - 1
          cap.set(1, currentFrameNum)
          ret, frame = cap.read()
      else:
        frame = back.copy()

    if self._hyperparameters["invertBlackWhiteOnImages"]:
      frame = 255 - frame

    if self._hyperparameters["imagePreProcessMethod"]:
      frame = preprocessImage(frame, self._hyperparameters)

    grey = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]
    initialCurFrame = curFrame.copy()

    # if False:
    putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )

    curFrame[putToWhite] = 255
    previousNbBlackPixels = []
    if self._hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"]:
      minPixel2nbBlackPixels = {}
      countTries = 0
      nbBlackPixels = 0
      nbBlackPixelsMax = int(self._hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"])
      while (minPixelDiffForBackExtract > 0) and (countTries < 30) and not(minPixelDiffForBackExtract in minPixel2nbBlackPixels):
        if countTries > 0:
          curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]
          putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
          curFrame[putToWhite] = 255
        ret, thresh1 = cv2.threshold(curFrame, self._hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
        thresh1 = 255 - thresh1
        nbBlackPixels = cv2.countNonZero(thresh1)
        minPixel2nbBlackPixels[minPixelDiffForBackExtract] = nbBlackPixels
        if nbBlackPixels > nbBlackPixelsMax:
          minPixelDiffForBackExtract = minPixelDiffForBackExtract + 1
        if nbBlackPixels <= nbBlackPixelsMax:
          minPixelDiffForBackExtract = minPixelDiffForBackExtract - 1
        countTries = countTries + 1
        previousNbBlackPixels.append(nbBlackPixels)
        if len(previousNbBlackPixels) >= 3:
          lastThree = previousNbBlackPixels[len(previousNbBlackPixels)-3: len(previousNbBlackPixels)]
          if lastThree.count(lastThree[0]) == len(lastThree):
            countTries = 1000000

      best_minPixelDiffForBackExtract = 0
      minDist = 10000000000000
      for minPixelDiffForBackExtract in minPixel2nbBlackPixels:
        nbBlackPixels = minPixel2nbBlackPixels[minPixelDiffForBackExtract]
        dist = abs(nbBlackPixels - nbBlackPixelsMax)
        if dist < minDist:
          minDist = dist
          best_minPixelDiffForBackExtract = minPixelDiffForBackExtract

      minPixelDiffForBackExtract = best_minPixelDiffForBackExtract
      self._hyperparameters["minPixelDiffForBackExtractHead"] = minPixelDiffForBackExtract
      curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]
      putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
      curFrame[putToWhite] = 255

    # else:
      # putToBlack3a = ( curFrame.astype('int32') <= (back.astype('int32') - minPixelDiffForBackExtract) )
      # putToBlack3b = ( curFrame.astype('int32') >= (back.astype('int32') + minPixelDiffForBackExtract) )

      # putToWhite  = (curFrame <= 220)
      # putToBlack  = (curFrame >= 220)

      # curFrame[putToWhite]   = 255
      # curFrame[putToBlack]   = 0
      # curFrame[putToBlack3a] = 0
      # curFrame[putToBlack3b] = 0

    if debug:
      self._debugFrame(frame, title='Frame')

    return [curFrame, initialCurFrame, back]

  def getImageSequential(self, cap, frameNumber, wellNumber):
    minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
    debug = 0

    xtop = self._wellPositions[wellNumber]['topLeftX']
    ytop = self._wellPositions[wellNumber]['topLeftY']
    lenX = self._wellPositions[wellNumber]['lengthX']
    lenY = self._wellPositions[wellNumber]['lengthY']

    ret, frame = cap.read()

    while not(ret):
      print("WARNING: was not able to extract the frame", str(frameNumber),"in 'getImageSequential'")
      frameNumber = frameNumber - 1
      cap.set(1, frameNumber)
      ret, frame = cap.read()

    if self._hyperparameters["invertBlackWhiteOnImages"]:
      frame = 255 - frame

    if self._hyperparameters["imagePreProcessMethod"]:
      frame = preprocessImage(frame, self._hyperparameters)

    grey = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]

    if debug:
      self._debugFrame(curFrame, title='Frame')

    return curFrame

  def headEmbededFrame(self, frameNumber, wellNumber):
    debug = 0

    cap = zzVideoReading.VideoCapture(self._videoPath)

    cap.set(1, frameNumber)
    ret, frame = cap.read()

    while not(ret):
      print("WARNING: couldn't read the frameNumber", frameNumber, "for the video", self._hyperparameters["videoName"])
      frameNumber = frameNumber - 1
      cap.set(1, frameNumber)
      ret, frame = cap.read()

    xtop = self._wellPositions[wellNumber]['topLeftX']
    ytop = self._wellPositions[wellNumber]['topLeftY']
    lenX = self._wellPositions[wellNumber]['lengthX']
    lenY = self._wellPositions[wellNumber]['lengthY']
    frame = frame[ytop:ytop+lenY, xtop:xtop+lenX]

    if ("invertBlackWhiteOnImages" in self._hyperparameters) and self._hyperparameters["invertBlackWhiteOnImages"]:
      frame = 255 - frame

    if ("imagePreProcessMethod" in self._hyperparameters) and self._hyperparameters["imagePreProcessMethod"]:
      frame = preprocessImage(frame, self._hyperparameters)

    kernel = np.ones((8,8),np.float32)/25
    thres1  = cv2.filter2D(frame,-1,kernel)
    retval, thres1 = cv2.threshold(thres1, 80, 255, cv2.THRESH_BINARY)
    thres1 = 255 - thres1

    frame = 255 - frame

    if (debug):
      self._debugFrame(frame, title='thres1')
      self._debugFrame(thres1, title='thres1')

    frame  = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    thres1 = cv2.cvtColor(thres1, cv2.COLOR_BGR2GRAY)

    return [frame, thres1]

  def headEmbededFrameBackExtract(self, background, frameNumber, wellNumber):
    minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
    debug = 0

    cap = zzVideoReading.VideoCapture(self._videoPath)

    cap.set(1, frameNumber)
    ret, frame = cap.read()

    while not(ret):
      print("WARNING: couldn't read the frameNumber", frameNumber, "for the video", self._hyperparameters["videoName"])
      frameNumber = frameNumber - 1
      cap.set(1, frameNumber)
      ret, frame = cap.read()

    xtop = self._wellPositions[wellNumber]['topLeftX']
    ytop = self._wellPositions[wellNumber]['topLeftY']
    lenX = self._wellPositions[wellNumber]['lengthX']
    lenY = self._wellPositions[wellNumber]['lengthY']
    frame = frame[ytop:ytop+lenY, xtop:xtop+lenX]
    background = background[ytop:ytop+lenY, xtop:xtop+lenX]

    if ("invertBlackWhiteOnImages" in self._hyperparameters) and self._hyperparameters["invertBlackWhiteOnImages"]:
      frame = 255 - frame

    if ("imagePreProcessMethod" in self._hyperparameters) and self._hyperparameters["imagePreProcessMethod"]:
      frame = preprocessImage(frame, self._hyperparameters)

    kernel = np.ones((8,8),np.float32)/25
    thres1  = cv2.filter2D(frame,-1,kernel)
    retval, thres1 = cv2.threshold(thres1, 80, 255, cv2.THRESH_BINARY)

    frame  = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    thres1 = cv2.cvtColor(thres1, cv2.COLOR_BGR2GRAY)

    if debug:
      self._debugFrame(frame, title='thres1')

    putToWhite = ( frame.astype('int32') >= (background.astype('int32') + minPixelDiffForBackExtract) )
    # This puts the pixels that belong to a fish to white
    frame[putToWhite] = 255

    if debug:
      self._debugFrame(frame, title='thres1')
      # cv2.imshow('thres1', thres1)
      # cv2.waitKey(0)

    thres1 = 255 - thres1
    frame  = 255 - frame

    return [frame, thres1]

  def _headEmbededFrameSequential(self, cap, frameNumber, wellNumber):
    debug = 0

    ret, frame = cap.read()

    while not(ret):
      print("WARNING: couldn't read the frameNumber", frameNumber, "for the video", self._hyperparameters["videoName"])
      frameNumber = frameNumber - 1
      cap.set(1, frameNumber)
      ret, frame = cap.read()

    xtop = self._wellPositions[wellNumber]['topLeftX']
    ytop = self._wellPositions[wellNumber]['topLeftY']
    lenX = self._wellPositions[wellNumber]['lengthX']
    lenY = self._wellPositions[wellNumber]['lengthY']
    frame = frame[ytop:ytop+lenY, xtop:xtop+lenX]

    if ("invertBlackWhiteOnImages" in self._hyperparameters) and self._hyperparameters["invertBlackWhiteOnImages"]:
      frame = 255 - frame

    if ("imagePreProcessMethod" in self._hyperparameters) and self._hyperparameters["imagePreProcessMethod"]:
      frame = preprocessImage(frame, self._hyperparameters)

    kernel = np.ones((8,8),np.float32)/25
    thres1  = cv2.filter2D(frame,-1,kernel)
    retval, thres1 = cv2.threshold(thres1, 80, 255, cv2.THRESH_BINARY)
    thres1 = 255 - thres1

    frame = 255 - frame

    if debug:
      self._debugFrame(frame, title='thres1')
      self._debugFrame(thres1, title='thres1')

    frame  = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    thres1 = cv2.cvtColor(thres1, cv2.COLOR_BGR2GRAY)

    return [frame, thres1]

  def _headEmbededFrameSequentialBackExtract(self, cap, frameNumber, wellNumber):
    minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
    debug = 0

    ret, frame = cap.read()

    while not(ret):
      print("WARNING: couldn't read the frameNumber", frameNumber, "for the video", self._hyperparameters["videoName"])
      frameNumber = frameNumber - 1
      cap.set(1, frameNumber)
      ret, frame = cap.read()

    xtop = self._wellPositions[wellNumber]['topLeftX']
    ytop = self._wellPositions[wellNumber]['topLeftY']
    lenX = self._wellPositions[wellNumber]['lengthX']
    lenY = self._wellPositions[wellNumber]['lengthY']
    frame = frame[ytop:ytop+lenY, xtop:xtop+lenX]
    background = self._background[ytop:ytop+lenY, xtop:xtop+lenX]

    if ("invertBlackWhiteOnImages" in self._hyperparameters) and self._hyperparameters["invertBlackWhiteOnImages"]:
      frame = 255 - frame

    if ("imagePreProcessMethod" in self._hyperparameters) and self._hyperparameters["imagePreProcessMethod"]:
      frame = preprocessImage(frame, self._hyperparameters)

    kernel = np.ones((8,8),np.float32)/25
    thres1  = cv2.filter2D(frame,-1,kernel)
    retval, thres1 = cv2.threshold(thres1, 80, 255, cv2.THRESH_BINARY)

    frame  = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    thres1 = cv2.cvtColor(thres1, cv2.COLOR_BGR2GRAY)

    if debug:
      self._debugFrame(frame, title='thres1')

    putToWhite = ( frame.astype('int32') >= (background.astype('int32') + minPixelDiffForBackExtract) )
    # This puts the pixels that belong to a fish to white
    frame[putToWhite] = 255

    if (debug):
      self._debugFrame(frame, title='thres1')

    thres1 = 255 - thres1
    frame  = 255 - frame

    return [frame, thres1]

  def _getImages(self, cap, i, wellNumber, alreadyExtractedImage=0, trackingHeadTailAllAnimals=0):
    xHead = 0
    yHead = 0

    initialCurFrame = 0
    back = 0

    if self._hyperparameters["headEmbeded"]:
      if self._hyperparameters["headEmbededRemoveBack"] == 0:
        if self._hyperparameters["adjustHeadEmbededTracking"] == 0 and not self._hyperparameters["adjustHeadEmbeddedEyeTracking"]:
          [frame, thresh1] = self._headEmbededFrameSequential(cap, i, wellNumber)
        else:
          [frame, thresh1] = self.headEmbededFrame(i, wellNumber)
        gray = frame.copy()
      else:
        if self._hyperparameters["adjustHeadEmbededTracking"] == 0 and not self._hyperparameters["adjustHeadEmbeddedEyeTracking"]:
          [frame, thresh1] = self._headEmbededFrameSequentialBackExtract(cap, i, wellNumber)
        else:
          [frame, thresh1] = self.headEmbededFrameBackExtract(self._background, i, wellNumber)
        gray = frame.copy()
    else:
      if self._hyperparameters["adjustFreelySwimTracking"] == 0 and self._hyperparameters["adjustFreelySwimTrackingAutomaticParameters"] == 0:
        if len(self._background):
          [frame, initialCurFrame, back, xHead, yHead] = self._getForegroundImageSequential(cap, i, wellNumber, alreadyExtractedImage, trackingHeadTailAllAnimals)
        else:
          frame = self.getImageSequential(cap, i, wellNumber)
      else:
        if len(self._background):
          [frame, initialCurFrame, back] = self.getForegroundImage(self._background, i, wellNumber)
        else:
          frame = self._getImage(i, wellNumber)

      gray = frame.copy()

      ret,thresh1 = cv2.threshold(gray,self._hyperparameters["thresholdForBlobImg"],255,cv2.THRESH_BINARY)

    coverHorizontalPortionBelowForHeadDetect = self._hyperparameters["coverHorizontalPortionBelowForHeadDetect"]
    if coverHorizontalPortionBelowForHeadDetect != -1:
      gray[coverHorizontalPortionBelowForHeadDetect:len(gray)-1,:] = 255

    coverHorizontalPortionAboveForHeadDetect = self._hyperparameters["coverHorizontalPortionAboveForHeadDetect"]
    if coverHorizontalPortionAboveForHeadDetect != -1:
      gray[0:coverHorizontalPortionAboveForHeadDetect,:] = 255

    coverVerticalPortionRightForHeadDetect = self._hyperparameters["coverVerticalPortionRightForHeadDetect"]
    if coverVerticalPortionRightForHeadDetect != -1:
      gray[:,coverVerticalPortionRightForHeadDetect:len(gray[0])-1] = 255

    coverPortionForHeadDetect = self._hyperparameters["coverPortionForHeadDetect"]
    if coverPortionForHeadDetect == "Bottom":
      gray[int(len(gray)/2):len(gray)-1, :] = 255
    if coverPortionForHeadDetect == "Top":
      gray[0:int(len(gray)/2), :] = 255
    if coverPortionForHeadDetect == "Left":
      gray[:, 0:int(len(gray[0])/2)] = 255
    if coverPortionForHeadDetect == "Right":
      gray[:, int(len(gray[0])/2):len(gray[0])-1] = 255

    if self._hyperparameters["debugCoverHorizontalPortionBelow"] and i==self._firstFrame:
      self._debugFrame(gray, title='Frame')

    # paramGaussianBlur = int((self._hyperparameters["paramGaussianBlur"] / 2)) * 2 + 1
    # blur = cv2.GaussianBlur(gray, (paramGaussianBlur, paramGaussianBlur), 0)
    blur = 0

    frame2 = frame #frame.copy()

    if self._hyperparameters["erodeIter"] or self._hyperparameters["dilateIter"]:
      erodeSize = self._hyperparameters["erodeSize"]
      kernel  = np.ones((erodeSize,erodeSize), np.uint8)

    if self._hyperparameters["erodeIter"]:
      thresh1 = cv2.erode(thresh1, kernel, iterations=self._hyperparameters["erodeIter"])

    if self._hyperparameters["dilateIter"]:
      thresh2 = thresh1.copy()
      thresh2 = cv2.dilate(thresh2, kernel, iterations=self._hyperparameters["dilateIter"])
    else:
      thresh2 = thresh1

    return [frame, gray, thresh1, blur, thresh2, frame2, initialCurFrame, back, xHead, yHead]



# from win32api import GetSystemMetrics
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import json
import numpy as np
import sys
import os
from pathlib import Path

def outputValidationVideo(videoPath, folderName, configFilePath, numWell, numBout, zoom, start, boutEnd, out, length, analyzeAllWellsAtTheSameTime, ZZoutputLocation=''):
  
  s1  = videoPath
  s2  = folderName
  s3  = "/"
  s3b = "results_"
  s4  = folderName
  s5  = ".avi"
  s5b = ".txt"
  
  cur_dir_path = os.path.dirname(os.path.realpath(__file__))
  initialPath  = Path(cur_dir_path)
  initialPath  = os.path.join(initialPath.parent.parent, 'ZZoutput')
  initialPath  = os.path.join(initialPath, s1)
  if len(ZZoutputLocation):
    initialPath = ZZoutputLocation
  
  videoPath   = os.path.join(os.path.join(initialPath, s2),       s4 + s5)
  resultsPath = os.path.join(os.path.join(initialPath, s2), s3b + s4 + s5b)
  
  cap = zzVideoReading.VideoCapture(videoPath)
  
  nx    = int(cap.get(3))
  ny    = int(cap.get(4))
  max_l = int(cap.get(7))
  
  with open(resultsPath) as f:
    supstruct = json.load(f)
  
  # Getting the information about the head positions
  topLeftX = supstruct["wellPositions"][numWell]["topLeftX"]
  topLeftY = supstruct["wellPositions"][numWell]["topLeftY"]
  HeadX = [pos + topLeftX for pos in supstruct["wellPoissMouv"][numWell][0][numBout]['HeadX']]
  HeadY = [pos + topLeftY for pos in supstruct["wellPoissMouv"][numWell][0][numBout]['HeadY']]
  
  if not("firstFrame" in supstruct):
    supstruct["firstFrame"] = 1
  
  boutStart = supstruct["wellPoissMouv"][numWell][0][numBout]['BoutStart']
  boutEnd   = supstruct["wellPoissMouv"][numWell][0][numBout]['BoutEnd']
  
  l = boutStart - supstruct["firstFrame"]
  
  while (l < boutEnd):
    
    cap.set(1, l )
    ret, img = cap.read()
    
    if ret and l - boutStart + supstruct["firstFrame"] < len(HeadX) and (numWell != -1) and (zoom):
        
      xmin = int(HeadX[l - boutStart + supstruct["firstFrame"]] - length/2)
      xmax = int(HeadX[l - boutStart + supstruct["firstFrame"]] + length/2)
      ymin = int(HeadY[l - boutStart + supstruct["firstFrame"]] - length/2)
      ymax = int(HeadY[l - boutStart + supstruct["firstFrame"]] + length/2)
      
      if (xmin < 0):
        xmin = 0
      if (ymin < 0):
        ymin = 0
      if (xmax > nx-1):
        xmax = nx-1
      if (ymax > ny-1):
        ymax = ny-1
      
      blank = np.zeros((length, length, 3), np.uint8)
    
      blank[0:ymax-ymin, 0:xmax-xmin] = img[ymin:ymax, xmin:xmax]
      
      out.write(blank)
    
    l = l + 1
  
  blank = np.zeros((length, length, 3), np.uint8)
  for k in range(0, 10):
    out.write(blank)
  
  return out



import zebrazoom
import pickle

# Applying the clustering on this dataframe

clusteringOptions = {
  'analyzeAllWellsAtTheSameTime' : 0, # put this to 1 for head-embedded videos, and to 0 for multi-well videos
  'pathToVideos' : '../ZZoutput/',
  'nbCluster' : 3,
  'modelUsedForClustering' : 'KMeans', # put either 'KMeans' or 'GaussianMixture' here
  #'nbPcaComponents' : 30,
  'scaleGraphs' : True,
  'showFigures' : False,
  'useFreqAmpAsym' : False,
  'useAngles' : False,
  'useAnglesSpeedHeadingDisp' : False,
  'useAnglesSpeedHeading' : True,
  'useAnglesSpeed' : False,
  'useAnglesHeading' : False,
  'useAnglesHeadingDisp' : False,
  'useFreqAmpAsymSpeedHeadingDisp' : False,
  'videoSaveFirstTenBouts' : False,
  'globalParametersCalculations' : True,
  'nbVideosToSave' : 10,
  'resFolder' : 'data',
  'nameOfFile' : 'example'
}

# Reloads the classifier previously created and get the nbFramesTakenIntoAccount value
infile = open('classifiers/classifier_'+clusteringOptions['nameOfFile']+'.txt','rb')
[classifier, nbFramesTakenIntoAccount] = pickle.load(infile)
infile.close()
clusteringOptions['nbFramesTakenIntoAccount'] = nbFramesTakenIntoAccount

# Classifies bouts based on the clustering previously obtained
[allBouts, c] = zebrazoom.applyClustering(clusteringOptions, classifier, 'resultsClustering')




from pathlib import Path
import copy
import json
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import os

import zebrazoom.code.paths as paths
from zebrazoom.code.vars import getGlobalVariables
globalVariables = getGlobalVariables()

cur_dir_path = os.path.dirname(os.path.realpath(__file__))

cur_dir_path2 = Path(cur_dir_path)
cur_dir_path2 = cur_dir_path2.parent

CONFIG_DEFAULT = \
{
  "videoFPS" : 0,
  "videoPixelSize" : 0,
  "firstFrameForTracking" : -1,
  "adjustDetectMovWithRawVideo" : 0,
  "adjustHeadEmbededTracking" : 0,
  "adjustFreelySwimTracking" : 0,
  "adjustFreelySwimTrackingAutomaticParameters" : 0,
  "adjustRectangularWellsDetect" : 0,
  "debugExtractParams" : 0,
  "debugTracking" : 0,
  "debugTrackingPtExtreme" : 0,
  "debugTrackingPtExtremeLargeVerticals" : 0,
  "debugTrackingThreshImg" : 0,
  "debugExtractBack" : 0,
  "debugFindWells" : 0,
  "debugHeadingCalculation" : 0,
  "debugValidationVideoHeading" : 0,
  "debugPauseBetweenTrackAndParamExtract" : "saveTrackDataAndExtractParam",
  "debugDetectMovWithRawVideo" : 0,
  "debugDetectMovWithRawVideoShowVid" : 1,
  "debugCoverHorizontalPortionBelow" : 0,
  "debugHeadEmbededFindNextPoints" : 0,
  "debugEyeTracking" : 0,
  "debugEyeTrackingAdvanced" : 0,
  "onlyDoTheTrackingForThisNumberOfFrames": 0,
  "createValidationVideo" : 1,
  "copyOriginalVideoToOutputFolderForValidation" : 0,
  "calculateAllTailAngles" : 0,
  "freqAlgoPosFollow" : 0,
  "popUpAlgoFollow" : 0,
  "closePopUpWindowAtTheEnd" : 1,
  "reloadWellPositions" : 0,
  "reloadWellPositionsFromFileInZZoutputIfItExistSaveInItOtherwise" : 0,
  "reloadBackground" : 0,
  "saveWellPositionsToBeReloadedNoMatterWhat" : 0,
  "backgroundExtractionForceUseAllVideoFrames" : 0,
  "updateBackgroundAtInterval" : 0,
  "useFirstFrameAsBackground" : 0,
  "exitAfterBackgroundExtraction" : 0,
  "exitAfterWellsDetection" : 0,
  "fasterMultiprocessing" : 0,
  "trackOnlyOnROI_halfDiameter" : 0,
  "tryCreatingFolderUntilSuccess" : 1,
  "searchPreviousFramesIfCurrentFrameIsCorrupted" : 1,
  "reduceImageResolutionPercentage" : 1,
  "trackingMethod" : "",
  "saveAllDataEvenIfNotInBouts" : 0,
  "savePathToOriginalVideoForValidationVideo" : 0,

  "setPixDiffBoutDetectParameters" : 0,

  "nbAnimalsPerWell" : 1,
  "trackTail" : 1,
  "multipleAnimalTrackingAdvanceAlgorithm" : 0,
  "multipleAnimalTrackingAdvanceAlgorithmMarginX" : 30,
  "multipleAnimalTrackingAdvanceAlgorithmDist1" : 50,
  "multipleAnimalTrackingAdvanceAlgorithmDist2" : 400,
  "findHeadPositionByUserInput" : 0,
  "accentuateFrameForManualTailExtremityFind" : 1,
  "outputFolder" : paths.getDefaultZZoutputFolder(),
  "onlyTrackThisOneWell" : -1,
  "noWellDetection" : 0,
  "multipleROIsDefinedDuringExecution" : 0,
  "oneWellManuallyChosenTopLeft" : [],
  "oneWellManuallyChosenBottomRight" : [],
  "headEmbeded" : 0,
  "headEmbededTeresaNicolson" : 0,
  "headEmbededRemoveBack" : 0,
  "extractBackWhiteBackground" : 1,
  "extractAdvanceZebraParameters" : 1,
  "takeTheHeadClosestToTheCenter" : 0,
  "backgroundSubtractorKNN" : 0,

  "groupOfMultipleSameSizeAndShapeEquallySpacedWells" : 0,
  "wellsAreRectangles" : 0,
  "rectangularWellMinMaxXandYmethod"       : 0,
  "rectangularWellMinMaxXandYmethodMargin" : 0,
  "findRectangleWellArea"                : 165000,
  "rectangleWellAreaImageThreshold"      : 105,
  "rectangleWellErodeDilateKernelSize"   : 20,
  "rectangularWellsInvertBlackWhite"     : 0,
  "rectangularWellStretchPercentage"     : 5,
  "rectangleWellAreaTolerancePercentage" : 20,

  "nbWells" : -1,
  "nbRowsOfWells"  : 0,
  "nbWellsPerRows" : 0,
  "coverHorizontalPortionBelowForHeadDetect" : -1,
  "coverHorizontalPortionAboveForHeadDetect" : -1,
  "coverVerticalPortionRightForHeadDetect"   : -1,
  "coverPortionForHeadDetect" : "Null",
  "centerOfMassTailTracking" : 0,
  "freeSwimmingTailTrackingMethod" : "tailExtremityDetect",
  "invertBlackWhiteOnImages" : 0,
  "generateAllTimeTailAngleGraph" : 0,
  "generateAllTimeTailAngleGraphLineWidth" : 0.3,

  "imagePreProcessMethod" : 0,
  "imagePreProcessParameters" : [],

  "backgroundPreProcessMethod" : 0,
  "backgroundPreProcessParameters" : [],

  "addBlackLineToImg_Width" : 0,

  "minWellDistanceForWellDetection" : 250,
  "wellOutputVideoDiameter" : -1,

  "nbImagesForBackgroundCalculation" : 60,
  "minPixelDiffForBackExtract" : 20,
  "adjustMinPixelDiffForBackExtract_nbBlackPixelsMax" : 0,
  "backgroundExtractionWithOnlyTwoFrames" : 0,
  "checkThatMovementOccurInVideo" : 0,
  "checkThatMovementOccurInVideoMedianFilterWindow" : 11,
  "setBackgroundToImageMedian" : 0,
  "firstFrameForBackExtract" : -1,
  "lastFrameForBackExtract"  : -1,

  "findContourPrecision" : "CHAIN_APPROX_SIMPLE",

  "paramGaussianBlur" : 31,
  "erodeSize"  : 3,
  "erodeIter"  : 1,
  "dilateIter" : 2,
  "thresholdForBlobImg" : 200,
  "multipleHeadTrackingIterativelyRelaxAreaCriteria" : 0,
  "minArea" : -1,
  "maxArea" : -1,
  "minAreaBody" : -1,
  "maxAreaBody" : -1,
  "headSize" : -1,
  "minTailSize" : 3,
  "maxTailSize" : 60,
  "tailExtremityMaxJugeDecreaseCoeff" : 0.3,
  "checkAllContourForTailExtremityDetect" : 0,
  "considerHighPointForTailExtremityDetect" : 1,
  "midlineIsInBlobTrackingOptimization" : 1,
  "nbTailPoints" : 10,
  "forceBlobMethodForHeadTracking" : 0,
  "postProcessMultipleTrajectories" : 0,
  "postProcessMaxDistanceAuthorized" : 100000000000000000000000000000000000000000000,
  "postProcessMaxDisapearanceFrames" : 100,
  "postProcessRemoveLowProbabilityDetection" : 0,
  "postProcessLowProbabilityDetectionThreshold" : 1,
  "postProcessLowProbabilityDetectionPercentOfMaximum" : 0,
  "postProcessRemovePointsOnBordersMargin" : 0,
  "postProcessRemovePointsAwayFromMainTrajectory" : 0,
  "postProcessRemovePointsAwayFromMainTrajectoryThreshold" : 2.5,
  "fixedHeadPositionX" : -1,
  "fixedHeadPositionY" : -1,
  "recalculateForegroundImageBasedOnBodyArea" : 0,
  "detectMouthInsteadOfHeadTwoSides" : 0,
  "findCenterOfAnimalByIterativelyDilating" : 0,
  "readjustCenterOfMassIfNotInsideContour": 0,

  "eyeTracking" : 0,
  "adjustHeadEmbeddedEyeTracking": 0,
  "invertColorsForHeadEmbeddedEyeTracking": 0,
  "improveContrastForEyeDetectionOfHeadEmbedded": 1,
  "eyeTrackingHeadEmbeddedWithSegment" : 0,
  "eyeTrackingHeadEmbeddedWithEllipse" : 0,
  "eyeFilterKernelSize" : 7,
  "headCenterToMidEyesPointDistance" : 5,
  "eyeBinaryThreshold" : 50,
  "midEyesPointToEyeCenterMaxDistance" : 10,
  "eyeHeadingSearchAreaHalfDiameter" : 40,
  "headingLineValidationPlotLength" : 10,
  "eyeTrackingHeadEmbeddedWidth" : 4,
  "eyeTrackingHeadEmbeddedWidthLeft"  : 0,
  "eyeTrackingHeadEmbeddedWidthRight" : 0,
  "eyeTrackingHeadEmbeddedHalfDiameter" : 15,

  "backCalculationStep" : -1,
  "step" : -1,
  "overwriteFirstStepValue" : 0,
  "overwriteLastStepValue"  : 0,
  "overwriteNbOfStepValues" : 1,
  "nbList" : -1,
  "expDecreaseFactor" : -1,
  "headEmbededMaxAngleBetweenSubsequentSegments" : 0,

  "automaticallySetSomeOfTheHeadEmbededHyperparameters" : 0,
  "headEmbededAutoSet_ExtendedDescentSearchOption" : 0,
  "headEmbededAutoSet_BackgroundExtractionOption"  : 0,
  "headEmbededParamInitialAngle" : 0.78539816339,
  "headEmbededParamGaussianBlur" : 13,
  "authorizedRelativeLengthTailEnd" : 0.85,
  "overwriteHeadEmbededParamGaussianBlur" : 0,
  "headEmbededParamTailDescentPixThreshStop" : 150,
  "headEmbededParamTailDescentPixThreshStopOverwrite" : -1,
  "headEmbededTailTrackFindMaxDepthInitialMaxDepth" : 300,
  "centerOfMassParamStep" : 25,
  "centerOfMassParamSegStep" : 5,
  "centerOfMassParamHalfDiam" : 13,
  "initialTailPortionMaxSegmentDiffAngleCutOffPos" : 0.15,
  "initialTailPortionMaxSegmentDiffAngleValue" : 1,
  "headEmbededRetrackIfWeirdInitialTracking": 0,

  "headingCalculationMethod" : "calculatedWithHead",

  "thresAngleBoutDetect" : -1,
  "windowForBoutDetectWithAngle" : 10,
  "thresForDetectMovementWithRawVideo" : 0,
  "minNbPixelForDetectMovementWithRawVideo" : 0,
  "frameGapComparision" : 1,
  "halfDiameterRoiBoutDetect" : 100,
  "fillGapFrameNb" : 5,
  "detectMovementWithRawVideoInsideTracking" : 0,
  "addBlackCircleOfHalfDiamOnHeadForBoutDetect" : 0,
  "boutEdgesWhereZeros" : 0,
  "noBoutsDetection" : 0,
  "addOneFrameAtTheEndForBoutDetection" : 0,
  "noPreProcessingOfImageForBoutDetection": 0,
  "coordinatesOnlyBoutDetection": 0,
  "coordinatesOnlyBoutDetectionMinDist": 0,

  "boutsMinNbFrames" : 1,
  "noChecksForBoutSelectionInExtractParams" : 1,
  "detectBoutMinNbFrames" : 2,
  "detectBoutMinDist" : 4,
  "detectBoutMinAngleDiff" : -1,
  "minNbPeaksForBoutDetect" : 1,

  "tailAngleSmoothingFactor" : 0.001,
  "tailAngleMedianFilter" : 3,

  "minFirstBendValue" : -1,
  "minDiffBetweenSubsequentBendAmp" : 0.02,
  "windowForLocalBendMinMaxFind" : 1,
  "minProminenceForBendsDetect" : 0.01,
  "doubleCheckBendMinMaxStatus" : 1,
  "removeFirstSmallBend" : 0,

  "saveSuperStructToMatlab" : 0,
  "removeLargeInstantaneousDistanceData" : 0,

  "plotOnlyOneTailPointForVisu" : 0,
  "trackingPointSizeDisplay" : 1,
  "validationVideoPlotHeading" : 1,
  "validationVideoPlotAnimalNumber": 0,

  "perBoutOutput" : 0,
  "perBoutOutputVideoStartStopFrameMargin" : 0,
  "perBoutOutputYaxis" : 0,
  "smoothTailHeadEmbeded" : -1,
  "smoothTailHeadEmbededNbOfIterations" : 1,
  "alternativeCurvatureCalculation" : 0,
  "curvatureMedianFilterSmoothingWindow" : 3,
  "nbPointsToIgnoreAtCurvatureBeginning" : 0,
  "nbPointsToIgnoreAtCurvatureEnd" : 0,
  "colorMapCurvature"  : "BrBG",
  "saveCurvaturePlots" : 0,
  "saveTailAngleGraph" : 0,
  "saveSubVideo"       : 0,
  "saveCurvatureData"  : 0,
  "maxCurvatureValues"     : 0,
  "curvatureXaxisNbFrames" : 0,

  "tailAnglesHeatMap" : 0,
  "tailAnglesHeatMapNbPointsToTakeIntoAccount" : 8,

  "createPandasDataFrameOfParameters" : 0,
  "frameStepForDistanceCalculation" : 0,

  "additionalOutputFolder" : "",
  "additionalOutputFolderOverwriteIfAlreadyExist" : 0,
  "outputValidationVideoContrastImprovement" : 0,
  "outputValidationVideoContrastImprovementQuartile" : 0.01,
  "outputValidationVideoFps" : -1,

  "computeEyesHeadingPlot": 0,

  "saveBodyMask": 0,
  "bodyMask_addWhitePoints": 0,
  "bodyMask_saveAsLabelMeJsonFormat": 0,
  "bodyMask_saveAsPngMask": 0,
  "saveBodyMaskResampleContourNbPoints": 0,
  "dontDeleteOutputFolderIfAlreadyExist": 0,
  "bodyMask_saveDataForAllFrames": 0,

  "trackingDL": 0,
  "unet": 0,
  "trackingDLdicotomySearchOfOptimalBlobArea": 0,
  "applySimpleThresholdOnPredictedMask": 0,
  "simpleThresholdCheckMinForMaxCountour": 0,
  "applyQuantileInDLalgo": 0,

  "fishTailTrackingDifficultBackground": 0,

  "firstFrame": 1,
  "lastFrame": 0,
  "videoWidth": 0,
  "videoHeight": 0,
  "storeH5": 0
}


def getHyperparameters(configFile, videoName, videoPath, argv):
  hyperparameters = copy.deepcopy(CONFIG_DEFAULT)
  hyperparameters["videoName"], _ = os.path.splitext(videoName)

  cap = zzVideoReading.VideoCapture(videoPath) if videoPath else None
  if cap is None or not cap.isOpened():
    print("Error opening video stream or file in getConfig")
  else:  # get defaults from video
    hyperparameters["lastFrame"] = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    hyperparameters["videoWidth"] = int(cap.get(3))
    hyperparameters["videoHeight"] = int(cap.get(4))

  if isinstance(configFile, str):
    with open(configFile) as f:
      config = json.load(f)
  else:
    assert isinstance(configFile, dict)
    config = configFile

  if "parentConfigFiles" in config:
    for parentConfigFile in config["parentConfigFiles"]:
      with open(os.path.join(cur_dir_path2, parentConfigFile)) as f2:
        parentConfig = json.load(f2)
        parentConfig.update(config)
        config = parentConfig

  hyperparameters.update(copy.deepcopy(config))

  if hyperparameters["tailAnglesHeatMap"]:
    hyperparameters["calculateAllTailAngles"] = 1

  if argv and argv[0] != "getTailExtremityFirstFrame":
    overridenParameters = {param: argv[i+1] for i, param in enumerate(argv) if str(param) in hyperparameters}
    for param, value in overridenParameters.items():
      print("command line hyperparameter change:", param, value)

      if param not in {"debugPauseBetweenTrackAndParamExtract", "outputFolder", "coverPortionForHeadDetect", "freeSwimmingTailTrackingMethod",
                       "findContourPrecision", "headingCalculationMethod", "additionalOutputFolder"}:
        try:
          overridenParameters[param] = int(value)
        except ValueError:
          overridenParameters[param] = float(value)

    hyperparameters.update(overridenParameters)
    config.update(overridenParameters)

  if hyperparameters["setPixDiffBoutDetectParameters"] > 0:
    hyperparameters["debugPauseBetweenTrackAndParamExtract"] = "justExtractParamFromPreviousTrackData"
    hyperparameters["createValidationVideo"]                 = 1
    hyperparameters["reloadBackground"]                      = 1
    hyperparameters["debugDetectMovWithRawVideo"]            = 1
    hyperparameters["debugDetectMovWithRawVideoShowVid"]     = 1
    hyperparameters["reloadWellPositions"]                   = 1

  if hyperparameters["setPixDiffBoutDetectParameters"] == 2:
    hyperparameters["debugDetectMovWithRawVideoShowVid"]     = 0

  # if globalVariables["limitedVersion"] == 1:
    # limitNbFrames = 1000
    # if hyperparameters["lastFrame"] - hyperparameters["firstFrame"] > limitNbFrames:
      # hyperparameters["lastFrame"] = int(hyperparameters["firstFrame"]) + limitNbFrames

  return hyperparameters, config


def getHyperparametersSimple(configTemp):
  hyperparameters = copy.deepcopy(CONFIG_DEFAULT)
  hyperparameters.update(copy.deepcopy(configTemp))
  return hyperparameters




import cv2
import numpy as np
import queue
import os

from ._baseZebraZoom import BaseZebraZoomTrackingMethod


class BaseFasterMultiprocessing(BaseZebraZoomTrackingMethod):
  def __init__(self, videoPath, wellPositions, hyperparameters):
    self._videoPath = videoPath
    self._videoName = os.path.splitext(os.path.basename(videoPath))[0]
    self._background = None
    self._hyperparameters = hyperparameters
    self._wellPositions = wellPositions
    self._firstWell = 0 if self._hyperparameters["onlyTrackThisOneWell"] == -1 else self._hyperparameters["onlyTrackThisOneWell"]
    self._lastWell = self._hyperparameters["nbWells"] - 1 if self._hyperparameters["onlyTrackThisOneWell"] == -1 else self._hyperparameters["onlyTrackThisOneWell"]
    self._auDessusPerAnimalIdList = None
    self._firstFrame = self._hyperparameters["firstFrame"]
    self._lastFrame = self._hyperparameters["lastFrame"]
    self._nbTailPoints = self._hyperparameters["nbTailPoints"]

    self._trackingHeadTailAllAnimalsList = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1, self._nbTailPoints, 2))
                                            for _ in range(self._hyperparameters["nbWells"])]
    self._trackingHeadingAllAnimalsList = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1))
                                           for _ in range(self._hyperparameters["nbWells"])]


  def _detectMovementWithRawVideoInsideTracking(self, i, grey, previousFrames):
    if previousFrames is None:
      previousFrames = queue.Queue(self._hyperparameters["frameGapComparision"])
      # previousXYCoords = queue.Queue(self._hyperparameters["frameGapComparision"])
      self._auDessusPerAnimalIdList = [[np.zeros((self._lastFrame-self._firstFrame+1, 1)) for nbAnimalsPerWell in range(0, self._hyperparameters["nbAnimalsPerWell"])]
                                       for wellNumber in range(self._hyperparameters["nbWells"])]
    halfDiameterRoiBoutDetect = self._hyperparameters["halfDiameterRoiBoutDetect"]
    for animal_Id in range(self._hyperparameters["nbAnimalsPerWell"]):
      if previousFrames.full():
        previousFrame   = previousFrames.get()
        curFrame        = grey.copy()
        for wellNumber in range(0 if self._hyperparameters["onlyTrackThisOneWell"] == -1 else self._hyperparameters["onlyTrackThisOneWell"], self._hyperparameters["nbWells"] if self._hyperparameters["onlyTrackThisOneWell"] == -1 else self._hyperparameters["onlyTrackThisOneWell"] + 1):
          # previousXYCoord = previousXYCoords.get()
          headX = self._trackingHeadTailAllAnimalsList[wellNumber][animal_Id, i-self._firstFrame][0][0]
          headY = self._trackingHeadTailAllAnimalsList[wellNumber][animal_Id, i-self._firstFrame][0][1]
          xmin = headX - self._hyperparameters["halfDiameterRoiBoutDetect"]
          ymin = headY - self._hyperparameters["halfDiameterRoiBoutDetect"]
          xmax = xmin + 2 * self._hyperparameters["halfDiameterRoiBoutDetect"]
          ymax = ymin + 2 * self._hyperparameters["halfDiameterRoiBoutDetect"]
          lenX = self._wellPositions[wellNumber]['lengthX']
          lenY = self._wellPositions[wellNumber]['lengthY']
          if xmin < 0:
            xmin = 0
          if ymin < 0:
            ymin = 0
          if xmax > lenX - 1:
            xmax = lenX - 1
          if ymax > lenY - 1:
            ymax = lenY - 1
          if ymax < ymin:
            ymax = ymin + 2 * self._hyperparameters["halfDiameterRoiBoutDetect"]
          if xmax < xmin:
            xmax = xmin + 2 * self._hyperparameters["halfDiameterRoiBoutDetect"]
          if ( (xmin > lenX - 1) or (xmax < 0) ):
            xmin = 0
            xmax = 0 + lenX - 1
          if ( (ymin > lenY - 1) or (ymax < 0) ):
            ymin = 0
            ymax = 0 + lenY - 1
          xmin = int(xmin + self._wellPositions[wellNumber]['topLeftX'])
          xmax = int(xmax + self._wellPositions[wellNumber]['topLeftX'])
          ymin = int(ymin + self._wellPositions[wellNumber]['topLeftY'])
          ymax = int(ymax + self._wellPositions[wellNumber]['topLeftY'])
          # img22       = img2[ymin:ymax, xmin:xmax]
          # imgFuture22 = imgFuture2[ymin:ymax, xmin:xmax]
          # maxX = min(len(previousFrame[0]), len(curFrame[0]))
          # maxY = min(len(previousFrame), len(curFrame))
          # print("Av: aaa:", len(previousFrame))
          # print("Av: bbb:", len(previousFrame[0]))
          # print("1Av: aaa:", len(grey))
          # print("1Av: bbb:", len(grey[0]))
          subPreviousFrame = previousFrame[ymin:ymax, xmin:xmax].copy()
          subCurFrame      = curFrame[ymin:ymax, xmin:xmax].copy()
          # print("Aft: aaa:", len(previousFrame))
          # print("Aft: bbb:", len(previousFrame[0]))
          # if i > 250 and wellNumber == 3:
            # print("wellNumber:", wellNumber, "; headX, headY:", headX, headY, "; xmin, xmax, ymin, ymax:", xmin, xmax, ymin, ymax)
            # print("aaa:", len(previousFrame))
            # print("bbb:", len(previousFrame[0]))
            # self._debugFrame(subPreviousFrame, title='subPreviousFrame' + str(wellNumber))
            # self._debugFrame(subCurFrame,      title='subCurFrame' + str(wellNumber))

          # Possible optimization in the future: refine the ROI based on halfDiameterRoiBoutDetect !!!
          res = cv2.absdiff(subPreviousFrame, subCurFrame)
          ret, res = cv2.threshold(res,self._hyperparameters["thresForDetectMovementWithRawVideo"],255,cv2.THRESH_BINARY)
          totDiff = cv2.countNonZero(res)

          for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
            if totDiff > self._hyperparameters["minNbPixelForDetectMovementWithRawVideo"]:
              self._auDessusPerAnimalIdList[wellNumber][animalId][i-self._firstFrame] = 1
            else:
              self._auDessusPerAnimalIdList[wellNumber][animalId][i-self._firstFrame] = 0
      else:
        for wellNumber in range(0 if self._hyperparameters["onlyTrackThisOneWell"] == -1 else self._hyperparameters["onlyTrackThisOneWell"], self._hyperparameters["nbWells"] if self._hyperparameters["onlyTrackThisOneWell"] == -1 else self._hyperparameters["onlyTrackThisOneWell"] + 1):
          for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
            self._auDessusPerAnimalIdList[wellNumber][animalId][i-self._firstFrame] = 0
      previousFrames.put(grey)
      # previousXYCoords.put([xHead, yHead])
    return previousFrames



import contextlib
import os

import h5py

from zebrazoom.code.paths import getDefaultZZoutputFolder


def _findResultsFile(videoName):
  videoName, _ = os.path.splitext(videoName)
  if os.path.isabs(videoName):
    ZZoutputPath = os.path.dirname(videoName)
    videoName = os.path.basename(videoName)
  else:
    ZZoutputPath = getDefaultZZoutputFolder()
  resultsPath = os.path.join(ZZoutputPath, f'{videoName}.h5')
  if os.path.exists(resultsPath):  # exact match
    return resultsPath
  path = next(reversed(sorted(name for name in os.listdir(ZZoutputPath)
                              if os.path.splitext(name)[0][:-20] == videoName)), None)
  if path is None:
    raise ValueError(f'video {videoName} not found in the ZZoutput folder ({ZZoutputPath})')
  return os.path.join(ZZoutputPath, path)


@contextlib.contextmanager
def openResultsFile(videoName, mode):
  with h5py.File(_findResultsFile(videoName), mode) as results:
    yield results
